/* OpenPGP routines */
/* Written by David Madore <david.madore@ens.fr>. */
/* Public domain (2001/05/11) */
/* This version last modified 2001/05/13 */

/* See RFC 2440 */

/* === The header === */

/* Put this in openpgp.h if you don't like having everything in one
 * big file. */

#ifndef _DMADORE_OPENPGP_H
#define _DMADORE_OPENPGP_H

#define OPENPGP_HASH_MD5 1
#define OPENPGP_HASH_SHA1 2

#define OPENPGP_CIPHER_BLOWFISH128 4
#define OPENPGP_CIPHER_TWOFISH256 10

struct openpgp_handle {
  char p;
  char usebuf;
  unsigned char buf[1024];
  int bufpos;
  void (* callback) (unsigned char byte, void *user_data);
  void *user_data;
};

/* Start an OpenPGP stream. */
void OpenPGP_start (struct openpgp_handle *handle,
		    void (* callback) (unsigned char byte,
				       void *user_data),
		    void *user_data);

/* Feed a byte in an OpenPGP stream. */
void OpenPGP_feed (struct openpgp_handle *handle,
		   unsigned char byte);

/* Stop an OpenPGP stream. */
void OpenPGP_stop (struct openpgp_handle *handle);

/* Terminate a packet. */
void OpenPGP_end_packet (struct openpgp_handle *handle);

/* Create a Literal Data Packet (RFC 2440 section 5.9).  format
 * specifies the type of data (0x62 for binary data, 0x74 for
 * canonical text data); fname is the file name ("_CONSOLE" for
 * special "For Your Eyes Only" data); mtime is the modification time
 * in seconds since 1970/01/01 at 0 hours UTC, or 0 for present;
 * data_len is the promised data length, or -1 if this can't be
 * predicted. */
void OpenPGP_literal_data (struct openpgp_handle *handle,
			   char format,
			   const char fname[], int fname_len,
			   unsigned long mtime,
			   long data_len);

long predict_literal_data_len (int fname_len, long data_len);

/* Create a Symmetrically Encrypted Data Packet (RFC 2440 section
 * 5.7). */
void OpenPGP_symenc_data (struct openpgp_handle *handle,
			  long data_len);

long predict_symenc_data_len (long data_len);

/* Create a Symmetrically Encrypted Session Key Packet (RFC 2440
 * section 5.3).  If the session key is not encrypted in the packet,
 * put seskey_len=0. */
void OpenPGP_symenc_seskey (struct openpgp_handle *handle,
			    unsigned char algo,
			    const unsigned char s2k[], int s2k_len,
			    long seskey_len);

long predict_symenc_seskey_len (int s2k_len, long seskey_len);

/* ASCII armor (RFC 2440, section 6).  typestring is the string
 * describing the type of content (e.g. "PGP MESSAGE"). */

struct ascii_armor_handle {
  char typestring[128];
  unsigned long crc24;
  unsigned char d[3];
  int b;
  int x;
  void (* callback) (char ch, void *user_data);
  void *user_data;
};

void ASCII_Armor_start (struct ascii_armor_handle *handle,
			const char typestring[], const char *headers,
			void (* callback) (char ch, void *user_data),
			void *user_data);

void ASCII_Armor_feed (struct ascii_armor_handle *handle,
		       unsigned char byte);

void ASCII_Armor_stop (struct ascii_armor_handle *handle);

long predict_armor_len (const char typestring[], const char *headers,
			long data_len);

#endif /* not defined _DMADORE_OPENPGP_H */

/* === The implementation === */
#include <string.h>

void
OpenPGP_start (struct openpgp_handle *handle,
	       void (* callback) (unsigned char byte, void *user_data),
	       void *user_data)
{
  handle->callback = callback;
  handle->user_data = user_data;
  handle->p = 0;
}

static void
OpenPGP_push (struct openpgp_handle *handle,
	      unsigned char byte)
{
  handle->callback (byte, handle->user_data);
}

void
OpenPGP_feed (struct openpgp_handle *handle,
	      unsigned char byte)
{
  if ( handle->usebuf )
    {
      if ( handle->bufpos >= 1024 )
	{
	  int i;

	  OpenPGP_push (handle, 0xea);
	  for ( i=0 ; i<1024 ; i++ )
	    OpenPGP_push (handle, handle->buf[i]);
	  handle->bufpos = 0;
	}
      handle->buf[handle->bufpos++] = byte;
    }
  else
    OpenPGP_push (handle, byte);
}

void
OpenPGP_stop (struct openpgp_handle *handle)
{
  if ( handle-> p )
    OpenPGP_end_packet (handle);
}

static void
encode_len (long len, unsigned char lenbuf[5], int *lenlen)
{
  /* RFC 2440 section 4.2.2 */
  if ( len < 192 )
    {
      lenbuf[0] = len;
      *lenlen = 1;
    }
  else if ( len < 8384 )
    {
      lenbuf[0] = ((len-192)>>8)+192;
      lenbuf[1] = (len-192)&0xff;
      *lenlen = 2;
    }
  else
    {
      lenbuf[0] = 0xff;
      lenbuf[1] = (len>>24)&0xff;
      lenbuf[2] = (len>>16)&0xff;
      lenbuf[3] = (len>>8)&0xff;
      lenbuf[4] = len&0xff;
      *lenlen = 5;
    }
}

static int
len_len (long len)
{
  if ( len < 192 )
    {
      return 1;
    }
  else if ( len < 8384 )
    {
      return 2;
    }
  else
    {
      return 5;
    }
}

static void
OpenPGP_write_len (struct openpgp_handle *handle,
		   long len)
{
  unsigned char lenbuf[5];
  int lenlen;
  int i;

  encode_len (len, lenbuf, &lenlen);
  for ( i=0 ; i<lenlen ; i++ )
    OpenPGP_push (handle, lenbuf[i]);
}

void
OpenPGP_end_packet (struct openpgp_handle *handle)
{
  int i;

  if ( handle->usebuf )
    {
      OpenPGP_write_len (handle, handle->bufpos);
      for ( i=0 ; i<handle->bufpos ; i++ )
	OpenPGP_push (handle, handle->buf[i]);
    }
  handle->p = 0;
}

static void
OpenPGP_start_len (struct openpgp_handle *handle,
		   long len)
{
  handle->p = 1;
  if ( len >= 0 )
    {
      handle->usebuf = 0;
      OpenPGP_write_len (handle, len);
    }
  else
    {
      handle->usebuf = 1;
      handle->bufpos = 0;
    }
}

void
OpenPGP_literal_data (struct openpgp_handle *handle,
		      char format,
		      const char fname[], int fname_len,
		      unsigned long mtime,
		      long data_len)
{
  long len;
  int i;

  if ( fname_len > 255 )
    fname_len = 255;
  len = data_len>=0 ? 1 + 1 + fname_len + 4 + data_len : -1;
  OpenPGP_push (handle, 0xcb);
  OpenPGP_start_len (handle, len);
  OpenPGP_feed (handle, format);
  OpenPGP_feed (handle, fname_len);
  for ( i=0 ; i<fname_len ; i++ )
    OpenPGP_feed (handle, fname[i]);
  OpenPGP_feed (handle, (mtime>>24)&0xf);
  OpenPGP_feed (handle, (mtime>>16)&0xf);
  OpenPGP_feed (handle, (mtime>>8)&0xf);
  OpenPGP_feed (handle, mtime&0xf);
}

long
predict_literal_data_len (int fname_len, long data_len)
{
  long len;

  len = 1 + 1 + fname_len + 4 + data_len;
  return 1 + len_len (len) + len;
}

void
OpenPGP_symenc_data (struct openpgp_handle *handle,
		     long data_len)
{
  long len;

  len = data_len;
  OpenPGP_push (handle, 0xc9);
  OpenPGP_start_len (handle, len);
}

long
predict_symenc_data_len (long data_len)
{
  long len;

  len = data_len;
  return 1 + len_len (len) + len;
}

void
OpenPGP_symenc_seskey (struct openpgp_handle *handle,
		       unsigned char algo,
		       const unsigned char s2k[], int s2k_len,
		       long seskey_len)
{
  long len;
  int i;

  len = seskey_len>=0 ? 2 + s2k_len + seskey_len : -1;
  OpenPGP_push (handle, 0xc3);
  OpenPGP_start_len (handle, len);
  OpenPGP_feed (handle, 0x4);
  OpenPGP_feed (handle, algo);
  for ( i=0 ; i<s2k_len ; i++ )
    OpenPGP_feed (handle, s2k[i]);
}

long
predict_symenc_seskey_len (int s2k_len, long seskey_len)
{
  long len;

  len = 2 + s2k_len + seskey_len;
  return 1 + len_len (len) + len;
}

static void
Armor_output_char (struct ascii_armor_handle *handle, char ch)
{
  handle->callback (ch, handle->user_data);
}

static void
Armor_output_string (struct ascii_armor_handle *handle,
		     const char str[])
{
  int i;

  for ( i=0 ; str[i] ; i++ )
    Armor_output_char (handle, str[i]);
}

void
ASCII_Armor_start (struct ascii_armor_handle *handle,
		   const char typestring[], const char *headers,
		   void (* callback) (char ch, void *user_data),
		   void *user_data)
{
  static const unsigned long crc24_init = 0xb704ceUL;

  strncpy (handle->typestring, typestring, 127);
  handle->typestring[127] = 0;
  handle->crc24 = crc24_init;
  handle->d[0] = 0;
  handle->d[1] = 0;
  handle->d[2] = 0;
  handle->b = 0;
  handle->x = 0;
  handle->callback = callback;
  handle->user_data = user_data;
  Armor_output_string (handle, "-----BEGIN ");
  Armor_output_string (handle, handle->typestring);
  Armor_output_string (handle, "-----\n");
  if ( headers )
    Armor_output_string (handle, headers);
  Armor_output_char (handle, '\n');
}

static const char base64[64]
= "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char pad = '=';

static void
Armor_produce (struct ascii_armor_handle *handle, char finish)
{
  char tetrad[4];

  if ( handle->b == 3 || ( finish && handle->b > 0 ) )
    {
      tetrad[0] = base64[(handle->d[0]>>2)&0x3f];
      tetrad[1] = base64[((handle->d[0]&0x3)<<4)|((handle->d[1]>>4)&0xf)];
      tetrad[2] = base64[((handle->d[1]&0xf)<<2)|((handle->d[2]>>6)&0x3)];
      tetrad[3] = base64[handle->d[2]&0x3f];
      if ( handle->b <= 2 )
	tetrad[3] = pad;
      if ( handle->b <= 1 )
	tetrad[2] = pad;
      handle->d[0] = 0;
      handle->d[1] = 0;
      handle->d[2] = 0;
      handle->b = 0;
      Armor_output_char (handle, tetrad[0]);
      Armor_output_char (handle, tetrad[1]);
      Armor_output_char (handle, tetrad[2]);
      Armor_output_char (handle, tetrad[3]);
      handle->x += 4;
      if ( handle->x >= 60 )
	{
	  Armor_output_char (handle, '\n');
	  handle->x = 0;
	}
    }
}


void
ASCII_Armor_feed (struct ascii_armor_handle *handle,
		  unsigned char byte)
{
  static const unsigned long crc24_poly = 0x1864cfbUL;
  int j;

  handle->d[handle->b++] = byte;
  Armor_produce (handle, 0);
  handle->crc24 ^= ((unsigned long)byte)<<16;
  for ( j=0 ; j<8 ; j++ )
    {
      handle->crc24 <<= 1;
      if ( handle->crc24 & 0x1000000UL )
	handle->crc24 ^= crc24_poly;
    }
}

void
ASCII_Armor_stop (struct ascii_armor_handle *handle)
{
  Armor_produce (handle, 1);
  if ( handle->x > 0 )
    {
      Armor_output_char (handle, '\n');
      handle->x = 0;
    }
  Armor_output_char (handle, pad);
  Armor_output_char (handle, base64[(handle->crc24>>18)&0x3f]);
  Armor_output_char (handle, base64[(handle->crc24>>12)&0x3f]);
  Armor_output_char (handle, base64[(handle->crc24>>6)&0x3f]);
  Armor_output_char (handle, base64[handle->crc24&0x3f]);
  Armor_output_char (handle, '\n');
  Armor_output_string (handle, "-----END ");
  Armor_output_string (handle, handle->typestring);
  Armor_output_string (handle, "-----\n");
}

long
predict_armor_len (const char typestring[], const char *headers,
		   long data_len)
{
  long typestring_len, len;

  typestring_len = strlen (typestring);
  if ( typestring_len > 127 )
    typestring_len = 127;
  len = data_len;
  len = ((len+2)/3)*4;  /* In base64 encoding */
  len += (len+59)/60;  /* With a LF every 60 chars */
  len += 5+1;  /* CRC24 (base64-encoded) + LF */
  len += 17+1+15;  /* "-----BEGIN -----\n"+"\n"+"-----END -----\n" */
  len += 2*typestring_len;
  if ( headers )
    len += strlen (headers);
  return len;
}
