/* OpenPGP routines */
/* Written by David Madore <david.madore@ens.fr>. */
/* Public domain (2001/05/11) */
/* This version last modified 2001/05/13 */

/* See RFC 2440 */

/* === The header === */

/* Put this in openpgp.h if you don't like having everything in one
 * big file. */

#ifndef _DMADORE_OPENPGP_H
#define _DMADORE_OPENPGP_H

#define OPENPGP_HASH_MD5 1
#define OPENPGP_HASH_SHA1 2

#define OPENPGP_CIPHER_BLOWFISH128 4
#define OPENPGP_CIPHER_TWOFISH256 10

struct openpgp_handle {
  char p;
  char usebuf;
  unsigned char buf[1024];
  int bufpos;
  void (* callback) (unsigned char byte, void *user_data);
  void *user_data;
};

/* Start an OpenPGP stream. */
void OpenPGP_start (struct openpgp_handle *handle,
		    void (* callback) (unsigned char byte,
				       void *user_data),
		    void *user_data);

/* Feed a byte in an OpenPGP stream. */
void OpenPGP_feed (struct openpgp_handle *handle,
		   unsigned char byte);

/* Stop an OpenPGP stream. */
void OpenPGP_stop (struct openpgp_handle *handle);

/* Terminate a packet. */
void OpenPGP_end_packet (struct openpgp_handle *handle);

/* Create a Literal Data Packet (RFC 2440 section 5.9).  format
 * specifies the type of data (0x62 for binary data, 0x74 for
 * canonical text data); fname is the file name ("_CONSOLE" for
 * special "For Your Eyes Only" data); mtime is the modification time
 * in seconds since 1970/01/01 at 0 hours UTC, or 0 for present;
 * data_len is the promised data length, or -1 if this can't be
 * predicted. */
void OpenPGP_literal_data (struct openpgp_handle *handle,
			   char format,
			   const char fname[], int fname_len,
			   unsigned long mtime,
			   long data_len);

long predict_literal_data_len (int fname_len, long data_len);

/* Create a Symmetrically Encrypted Data Packet (RFC 2440 section
 * 5.7). */
void OpenPGP_symenc_data (struct openpgp_handle *handle,
			  long data_len);

long predict_symenc_data_len (long data_len);

/* Create a Symmetrically Encrypted Session Key Packet (RFC 2440
 * section 5.3).  If the session key is not encrypted in the packet,
 * put seskey_len=0. */
void OpenPGP_symenc_seskey (struct openpgp_handle *handle,
			    unsigned char algo,
			    const unsigned char s2k[], int s2k_len,
			    long seskey_len);

long predict_symenc_seskey_len (int s2k_len, long seskey_len);

/* ASCII armor (RFC 2440, section 6).  typestring is the string
 * describing the type of content (e.g. "PGP MESSAGE"). */

struct ascii_armor_handle {
  char typestring[128];
  unsigned long crc24;
  unsigned char d[3];
  int b;
  int x;
  void (* callback) (char ch, void *user_data);
  void *user_data;
};

void ASCII_Armor_start (struct ascii_armor_handle *handle,
			const char typestring[], const char *headers,
			void (* callback) (char ch, void *user_data),
			void *user_data);

void ASCII_Armor_feed (struct ascii_armor_handle *handle,
		       unsigned char byte);

void ASCII_Armor_stop (struct ascii_armor_handle *handle);

long predict_armor_len (const char typestring[], const char *headers,
			long data_len);

#endif /* not defined _DMADORE_OPENPGP_H */

/* EOF */
