/* Twofish routines, after Bruce Schneier &al */
/* Written by David Madore <david.madore@ens.fr>. */
/* Public domain (2001-05-11) */
/* This version last modified 2003-02-16 */

/* Note: these routines do not depend on endianness. */

/* === The header === */

/* Put this in twofish.h if you don't like having everything in one
 * big file. */

#ifndef _DMADORE_TWOFISH_H
#define _DMADORE_TWOFISH_H

/* --- Basic twofish routines --- */

#define TWOFISH_NBROUNDS 16

struct twf_ctx {
  /* The subkeys used by the twofish cipher */
  unsigned long K[2*TWOFISH_NBROUNDS+8];
  unsigned long S[4][256];
};

/* Encipher one 128-bit quantity (divided in four 32-bit quantities)
 * using the precalculated subkeys). */
void Twofish_encipher (const struct twf_ctx *c,
		       unsigned long *x0, unsigned long *x1,
		       unsigned long *x2, unsigned long *x3);

/* Decipher one 128-bit quantity (divided in four 32-bit quantities)
 * using the precalculated subkeys). */
void Twofish_decipher (const struct twf_ctx *c,
		       unsigned long *x0, unsigned long *x1,
		       unsigned long *x2, unsigned long *x3);

/* Initialize the cipher by calculating the subkeys from the key. */
void Twofish_initialize (struct twf_ctx *c,
			 const unsigned char key[], int key_bytes);

/* --- Twofish used in Electronic Code Book (ECB) mode --- */

struct twf_ecb_ctx {
  /* Whether we are encrypting (rather than decrypting) */
  char encrypt;
  /* The twofish subkeys */
  struct twf_ctx c;
  /* The 128 bits of data being written */
  unsigned long d0, d1, d2, d3;
  /* Our position within the 128 bits (always between 0 and 15) */
  int b;
  /* The callback function to be called with every byte produced */
  void (* callback) (unsigned char byte, void *user_data);
  /* The user data to pass the the callback function */
  void *user_data;
};

/* Start an ECB Twofish cipher session: specify whether we are
 * encrypting or decrypting, what key is to be used, and what callback
 * should be called for every byte produced. */
void Twofish_ecb_start (struct twf_ecb_ctx *c, char encrypt,
			const unsigned char key[], int key_bytes,
			void (* callback) (unsigned char byte,
					   void *user_data),
			void *user_data);

/* Feed one byte to an ECB Twofish cipher session. */
void Twofish_ecb_feed (struct twf_ecb_ctx *c, unsigned char inb);

/* Stop an ECB Twofish session (i.e. flush the remaining bytes). */
void Twofish_ecb_stop (struct twf_ecb_ctx *c);

/* --- Twofish used in Cipher Block Chaining (CBC) mode --- */

struct twf_cbc_ctx {
  /* Whether we are encrypting (rather than decrypting) */
  char encrypt;
  /* The twofish subkeys */
  struct twf_ctx c;
  /* The 128 bits of data being written */
  unsigned long d0, d1, d2, d3;
  /* The last 128 bits of ciphertext */
  unsigned long l0, l1, l2, l3;
  /* Our position within the 128 bits (always between 0 and 15) */
  int b;
  /* The callback function to be called with every byte produced */
  void (* callback) (unsigned char byte, void *user_data);
  /* The user data to pass the the callback function */
  void *user_data;
};

/* Start a CBC Twofish cipher session: specify whether we are
 * encrypting or decrypting, what key is to be used, and what callback
 * should be called for every byte produced. */
void Twofish_cbc_start (struct twf_cbc_ctx *c, char encrypt,
			const unsigned char key[], int key_bytes,
			void (* callback) (unsigned char byte,
					   void *user_data),
			void *user_data);

/* Ditto, specifying Initialization Vector. */
void Twofish_cbc_start_iv (struct twf_cbc_ctx *c, char encrypt,
			   const unsigned char key[], int key_bytes,
			   unsigned long iv0, unsigned long iv1,
			   unsigned long iv2, unsigned long iv3,
			   void (* callback) (unsigned char byte,
					      void *user_data),
			   void *user_data);

/* Feed one byte to a CBC Twofish cipher session. */
void Twofish_cbc_feed (struct twf_cbc_ctx *c, unsigned char inb);

/* Stop a CBC Twofish session (i.e. flush the remaining bytes). */
void Twofish_cbc_stop (struct twf_cbc_ctx *c);

/* --- Twofish used in Cipher Feedback (CFB) mode --- */

struct twf_cfb_ctx {
  /* Whether we are encrypting (rather than decrypting) */
  char encrypt;
  /* The twofish subkeys */
  struct twf_ctx c;
  /* The 128-bit feedback register */
  unsigned long fr0, fr1, fr2, fr3;
  /* The 128-bit encrypted feedback register */
  unsigned long fre0, fre1, fre2, fre3;
  /* Our position within the 128 bits (always between 0 and 15) */
  int b;
  /* The callback function to be called with every byte produced */
  void (* callback) (unsigned char byte, void *user_data);
  /* The user data to pass the the callback function */
  void *user_data;
};

/* Start a CFB Twofish cipher session: specify whether we are
 * encrypting or decrypting, what key is to be used, and what callback
 * should be called for every byte produced. */
void Twofish_cfb_start (struct twf_cfb_ctx *c, char encrypt,
			const unsigned char key[], int key_bytes,
			void (* callback) (unsigned char byte,
					   void *user_data),
			void *user_data);

/* Ditto, specifying Initialization Vector. */
void Twofish_cfb_start_iv (struct twf_cfb_ctx *c, char encrypt,
			   const unsigned char key[], int key_bytes,
			   unsigned long iv0, unsigned long iv1,
			   unsigned long iv2, unsigned long iv3,
			   void (* callback) (unsigned char byte,
					      void *user_data),
			   void *user_data);

/* Feed one byte to a CFB Twofish cipher session. */
void Twofish_cfb_feed (struct twf_cfb_ctx *c, unsigned char inb);

/* Stop a CFB Twofish session (i.e. flush the remaining bytes). */
void Twofish_cfb_stop (struct twf_cfb_ctx *c);

/* "Resync" a CFB Twofish session - this is used in OpenPGP (see RFC
 * 2440 section 12.8). */
void Twofish_cfb_resync (struct twf_cfb_ctx *c);

#endif /* not defined _DMADORE_TWOFISH_H */

/* === The implementation === */

static unsigned char
slow_gf8_multiply_1 (unsigned char a, unsigned char b)
     /* Multiply in GF(256) modulo the primitive polynomial
      * x^8+x^6+x^5+x^3+1 (used in MDS multiplication). */
{
  unsigned int bb;
  unsigned char c;

  c = 0;
  bb = b;
  while ( a )
    {
      if ( a & 1 )
	c ^= bb;
      a >>= 1;
      bb <<= 1;
      if ( bb & 0x100 )
	bb ^= 0x169;
    }
  return c;
}

static char have_mds_table;
static unsigned char mds_table[4][2][16][4];

static void
compute_mds_table (void)
{
  static unsigned char mds_coeffs[4][4] = {
    { 0x01, 0xef, 0x5b, 0x5b },
    { 0x5b, 0xef, 0xef, 0x01 },
    { 0xef, 0x5b, 0x01, 0xef },
    { 0xef, 0x01, 0xef, 0x5b }
  };
  int i, j, k;

  have_mds_table = 1;
  for ( i=0 ; i<4 ; i++ )
    {
      for ( j=0 ; j<16 ; j++ )
	for ( k=0 ; k<4 ; k++ )
	  mds_table[i][0][j][k]
	    = slow_gf8_multiply_1 (mds_coeffs[i][k], j);
      for ( j=0 ; j<16 ; j++ )
	for ( k=0 ; k<4 ; k++ )
	  mds_table[i][1][j][k]
	    = slow_gf8_multiply_1 (mds_coeffs[i][k], j<<4);
    }
}

static unsigned long
mds (unsigned long y)
{
  int i, j;
  unsigned long z;

  if ( ! have_mds_table )
    compute_mds_table ();
  z = 0;
  for ( i=0 ; i<4 ; i++ )
    {
      for ( j=0 ; j<4 ; j++ )
	z ^= ((unsigned long)mds_table[i][0][(y>>(8*j))&0xf][j])<<(8*i);
      for ( j=0 ; j<4 ; j++ )
	z ^= ((unsigned long)mds_table[i][1][(y>>(8*j+4))&0xf][j])<<(8*i);
    }
  return z;
}

static unsigned long
g (const struct twf_ctx *tc, unsigned long x)
{
  unsigned char a, b, c, d;
  unsigned long y;

  a = x & 0xff;
  x >>= 8;
  b = x & 0xff;
  x >>= 8;
  c = x & 0xff;
  x >>= 8;
  d = x & 0xff;
  y = tc->S[3][d];
  y <<= 8;
  y |= tc->S[2][c];
  y <<= 8;
  y |= tc->S[1][b];
  y <<= 8;
  y |= tc->S[0][a];
  return mds (y);
}

void
Twofish_encipher (const struct twf_ctx *c,
		  unsigned long *x0, unsigned long *x1,
		  unsigned long *x2, unsigned long *x3)
{
  unsigned long X0, X1, X2, X3, t0, t1, f0, f1, temp0, temp1;
  int i;

  X0 = *x0;  X1 = *x1;  X2 = *x2;  X3 = *x3;
  X0 ^= c->K[0];  X1 ^= c->K[1];  X2 ^= c->K[2];  X3 ^= c->K[3];
  for ( i=0 ; i<TWOFISH_NBROUNDS ; i++ )
    {
      t0 = g (c, X0);
      t1 = g (c, ((X1<<8)&0xffffffffUL) | (X1>>24));
      f0 = (t0+t1+c->K[2*i+8])&0xffffffffUL;
      f1 = (t0+2*t1+c->K[2*i+9])&0xffffffffUL;
      X3 = ((X3<<1)&0xffffffffUL) | (X3>>31);
      X2 ^= f0;  X3 ^= f1;
      X2 = (X2>>1) | ((X2&1)<<31);
      temp0 = X0;  temp1 = X1;
      X0 = X2;  X1 = X3;
      X2 = temp0;  X3 = temp1;
    }
  temp0 = X0;  temp1 = X1;
  X0 = X2;  X1 = X3;
  X2 = temp0;  X3 = temp1;
  X0 ^= c->K[4];  X1 ^= c->K[5];  X2 ^= c->K[6];  X3 ^= c->K[7];
  *x0 = X0;  *x1 = X1;  *x2 = X2;  *x3 = X3;
}

void
Twofish_decipher (const struct twf_ctx *c,
		  unsigned long *x0, unsigned long *x1,
		  unsigned long *x2, unsigned long *x3)
{
  unsigned long X0, X1, X2, X3, t0, t1, f0, f1, temp0, temp1;
  int i;

  X0 = *x0;  X1 = *x1;  X2 = *x2;  X3 = *x3;
  X0 ^= c->K[4];  X1 ^= c->K[5];  X2 ^= c->K[6];  X3 ^= c->K[7];
  for ( i=TWOFISH_NBROUNDS-1 ; i>=0 ; i-- )
    {
      t0 = g (c, X0);
      t1 = g (c, ((X1<<8)&0xffffffffUL) | (X1>>24));
      f0 = (t0+t1+c->K[2*i+8])&0xffffffffUL;
      f1 = (t0+2*t1+c->K[2*i+9])&0xffffffffUL;
      X2 = ((X2<<1)&0xffffffffUL) | (X2>>31);
      X2 ^= f0;  X3 ^= f1;
      X3 = (X3>>1) | ((X3&1)<<31);
      temp0 = X0;  temp1 = X1;
      X0 = X2;  X1 = X3;
      X2 = temp0;  X3 = temp1;
    }
  temp0 = X0;  temp1 = X1;
  X0 = X2;  X1 = X3;
  X2 = temp0;  X3 = temp1;
  X0 ^= c->K[0];  X1 ^= c->K[1];  X2 ^= c->K[2];  X3 ^= c->K[3];
  *x0 = X0;  *x1 = X1;  *x2 = X2;  *x3 = X3;
}

static unsigned char
slow_gf8_multiply_2 (unsigned char a, unsigned char b)
     /* Multiply in GF(256) modulo the primitive polynomial
      * x^8+x^6+x^3+x^2+1 (used in RS multiplication). */
{
  unsigned int bb;
  unsigned char c;

  c = 0;
  bb = b;
  while ( a )
    {
      if ( a & 1 )
	c ^= bb;
      a >>= 1;
      bb <<= 1;
      if ( bb & 0x100 )
	bb ^= 0x14d;
    }
  return c;
}

static char have_rs_table;
static unsigned char rs_table[4][2][16][8];

static void
compute_rs_table (void)
{
  static unsigned char rs_coeffs[4][8] = {
    { 0x01, 0xa4, 0x55, 0x87, 0x5a, 0x58, 0xdb, 0x9e },
    { 0xa4, 0x56, 0x82, 0xf3, 0x1e, 0xc6, 0x68, 0xe5 },
    { 0x02, 0xa1, 0xfc, 0xc1, 0x47, 0xae, 0x3d, 0x19 },
    { 0xa4, 0x55, 0x87, 0x5a, 0x58, 0xdb, 0x9e, 0x03 }
  };
  int i, j, k;

  have_rs_table = 1;
  for ( i=0 ; i<4 ; i++ )
    {
      for ( j=0 ; j<16 ; j++ )
	for ( k=0 ; k<8 ; k++ )
	  rs_table[i][0][j][k]
	    = slow_gf8_multiply_2 (rs_coeffs[i][k], j);
      for ( j=0 ; j<16 ; j++ )
	for ( k=0 ; k<8 ; k++ )
	  rs_table[i][1][j][k]
	    = slow_gf8_multiply_2 (rs_coeffs[i][k], j<<4);
    }
}

static unsigned long
rs (unsigned long y0, unsigned long y1)
{
  int i, j;
  unsigned long z;

  if ( ! have_rs_table )
    compute_rs_table ();
  z = 0;
  for ( i=0 ; i<4 ; i++ )
    {
      for ( j=0 ; j<4 ; j++ )
	z ^= ((unsigned long)rs_table[i][0][(y0>>(8*j))&0xf][j])<<(8*i);
      for ( j=0 ; j<4 ; j++ )
	z ^= ((unsigned long)rs_table[i][0][(y1>>(8*j))&0xf][j+4])<<(8*i);
      for ( j=0 ; j<4 ; j++ )
	z ^= ((unsigned long)rs_table[i][1][(y0>>(8*j+4))&0xf][j])<<(8*i);
      for ( j=0 ; j<4 ; j++ )
	z ^= ((unsigned long)rs_table[i][1][(y1>>(8*j+4))&0xf][j+4])<<(8*i);
    }
  return z;
}

static char have_q_table;
static unsigned char q_table[2][256];

static void
compute_q_table (void)
{
  static unsigned char q_coeffs[2][4][16] = {
    {
      { 0x8,0x1,0x7,0xd,0x6,0xf,0x3,0x2,0x0,0xb,0x5,0x9,0xe,0xc,0xa,0x4 },
      { 0xe,0xc,0xb,0x8,0x1,0x2,0x3,0x5,0xf,0x4,0xa,0x6,0x7,0x0,0x9,0xd },
      { 0xb,0xa,0x5,0xe,0x6,0xd,0x9,0x0,0xc,0x8,0xf,0x3,0x2,0x4,0x7,0x1 },
      { 0xd,0x7,0xf,0x4,0x1,0x2,0x6,0xe,0x9,0xb,0x3,0x0,0x8,0x5,0xc,0xa }
    },
    {
      { 0x2,0x8,0xb,0xd,0xf,0x7,0x6,0xe,0x3,0x1,0x9,0x4,0x0,0xa,0xc,0x5 },
      { 0x1,0xe,0x2,0xb,0x4,0xc,0x3,0x7,0x6,0xd,0xa,0x5,0xf,0x9,0x0,0x8 },
      { 0x4,0xc,0x7,0x5,0x1,0x6,0x9,0xa,0x0,0xe,0xd,0x8,0x2,0xb,0x3,0xf },
      { 0xb,0x9,0x5,0x1,0xc,0x3,0xd,0xe,0x6,0x4,0x7,0xf,0x2,0x0,0x8,0xa }
    }
  };
  int i, j;

  have_q_table = 1;
  for ( i=0 ; i<2 ; i++ )
    for ( j=0 ; j<256 ; j++ )
      {
	unsigned char a0,b0,a1,b1,a2,b2,a3,b3,a4,b4;

	a0 = j>>4;  b0 = j&0xf;
	a1 = a0^b0;  b1 = a0^((b0>>1)|((b0&0x1)<<3))^((a0<<3)&0x8);
	a2 = q_coeffs[i][0][a1];  b2 = q_coeffs[i][1][b1];
	a3 = a2^b2;  b3 = a2^((b2>>1)|((b2&0x1)<<3))^((a2<<3)&0x8);
	a4 = q_coeffs[i][2][a3];  b4 = q_coeffs[i][3][b3];
	q_table[i][j] = (b4<<4)|a4;
      }
}

static unsigned long
h_without_mds (int k, const unsigned long l[], unsigned long x)
{
  unsigned char y[4];
  int i;

  if ( ! have_q_table )
    compute_q_table ();
  for ( i=0 ; i<4 ; i++ )
    y[i] = (x>>(8*i))&0xff;
  if ( k > 3 )
    {
      y[0] = q_table[1][y[0]] ^ (l[3]&0xff);
      y[1] = q_table[0][y[1]] ^ ((l[3]>>8)&0xff);
      y[2] = q_table[0][y[2]] ^ ((l[3]>>16)&0xff);
      y[3] = q_table[1][y[3]] ^ ((l[3]>>24)&0xff);
    }
  if ( k > 2 )
    {
      y[0] = q_table[1][y[0]] ^ (l[2]&0xff);
      y[1] = q_table[1][y[1]] ^ ((l[2]>>8)&0xff);
      y[2] = q_table[0][y[2]] ^ ((l[2]>>16)&0xff);
      y[3] = q_table[0][y[3]] ^ ((l[2]>>24)&0xff);
    }
  if ( k > 1 ) /* Always true */
    {
      y[0] = q_table[0][y[0]] ^ (l[1]&0xff);
      y[1] = q_table[1][y[1]] ^ ((l[1]>>8)&0xff);
      y[2] = q_table[0][y[2]] ^ ((l[1]>>16)&0xff);
      y[3] = q_table[1][y[3]] ^ ((l[1]>>24)&0xff);
    }
  if ( k > 0 ) /* Always true */
    {
      y[0] = q_table[0][y[0]] ^ (l[0]&0xff);
      y[1] = q_table[0][y[1]] ^ ((l[0]>>8)&0xff);
      y[2] = q_table[1][y[2]] ^ ((l[0]>>16)&0xff);
      y[3] = q_table[1][y[3]] ^ ((l[0]>>24)&0xff);
    }
  y[0] = q_table[1][y[0]];
  y[1] = q_table[0][y[1]];
  y[2] = q_table[1][y[2]];
  y[3] = q_table[0][y[3]];
  return ((unsigned long)y[0]) | (((unsigned long)y[1])<<8)
    | (((unsigned long)y[2])<<16) | (((unsigned long)y[3])<<24);
}

void
Twofish_initialize (struct twf_ctx *c,
		     const unsigned char key[], int key_bytes)
{
  int i,j,k;
  unsigned long key_words[8];
  unsigned long s[4];
  unsigned long m0[4];
  unsigned long m1[4];
  unsigned long a,b,z;

  if ( key_bytes > 24 )
    k = 4;
  else if ( key_bytes > 16 )
    k = 3;
  else k = 2;
  for ( i=0 ; i<2*k ; i++ )
    {
      key_words[i] = 0;
      for ( j=0 ; j<4 ; j++ )
	if ( i*4+j < key_bytes )
	  key_words[i] |= ((unsigned long)key[i*4+j])<<(j*8);
    }
  for ( i=0 ; i<k ; i++ )
    {
      m0[i] = key_words[2*i];
      m1[i] = key_words[2*i+1];
      s[k-1-i] = rs (key_words[2*i], key_words[2*i+1]);
    }
  for ( i=0 ; i<256 ; i++ )
    {
      a = i | (i<<8) | (i<<16) | (i<<24);
      z = h_without_mds (k, s, a);
      for ( j=0 ; j<4 ; j++ )
	c->S[j][i] = (z>>(8*j))&0xff;
    }
  for ( i=0 ; i<TWOFISH_NBROUNDS+4 ; i++ )
    {
      a = (2*i) | ((2*i)<<8) | ((2*i)<<16) | ((2*i)<<24);
      b = (2*i+1) | ((2*i+1)<<8) | ((2*i+1)<<16) | ((2*i+1)<<24);
      a = mds (h_without_mds (k, m0, a));
      b = mds (h_without_mds (k, m1, b));
      b = ((b<<8)&0xffffffffUL) | (b>>24);
      c->K[2*i] = (a+b)&0xffffffffUL;
      z = (a+2*b)&0xffffffffUL;
      c->K[2*i+1] = ((z<<9)&0xffffffffUL) | (z>>23);
    }
}

void
Twofish_ecb_start (struct twf_ecb_ctx *c, char encrypt,
		    const unsigned char key[], int key_bytes,
		    void (* callback) (unsigned char byte, void *user_data),
		    void *user_data)
{
  c->encrypt = encrypt;
  Twofish_initialize (&c->c, key, key_bytes);
  c->b = 0;
  c->d0 = 0;  c->d1 = 0;  c->d2 = 0;  c->d3 = 0;
  c->callback = callback;
  c->user_data = user_data;
}

void
Twofish_ecb_feed (struct twf_ecb_ctx *c, unsigned char inb)
{
  if ( c->b < 4 )
    c->d0 |= ((unsigned long)inb)<<(8*c->b);
  else if ( c->b < 8 )
    c->d1 |= ((unsigned long)inb)<<(8*(c->b-4));
  else if ( c->b < 12 )
    c->d2 |= ((unsigned long)inb)<<(8*(c->b-8));
  else
    c->d3 |= ((unsigned long)inb)<<(8*(c->b-12));
  c->b++;
  if ( c->b >= 16 )
    /* We have one block of data */
    {
      unsigned long x0, x1, x2, x3;

      x0 = c->d0;  x1 = c->d1;  x2 = c->d2;  x3 = c->d3;
      (c->encrypt ? Twofish_encipher : Twofish_decipher) (&c->c,
							  &x0, &x1, &x2, &x3);
      c->callback (x0&0xff, c->user_data);
      c->callback ((x0>>8)&0xff, c->user_data);
      c->callback ((x0>>16)&0xff, c->user_data);
      c->callback ((x0>>24)&0xff, c->user_data);
      c->callback (x1&0xff, c->user_data);
      c->callback ((x1>>8)&0xff, c->user_data);
      c->callback ((x1>>16)&0xff, c->user_data);
      c->callback ((x1>>24)&0xff, c->user_data);
      c->callback (x2&0xff, c->user_data);
      c->callback ((x2>>8)&0xff, c->user_data);
      c->callback ((x2>>16)&0xff, c->user_data);
      c->callback ((x2>>24)&0xff, c->user_data);
      c->callback (x3&0xff, c->user_data);
      c->callback ((x3>>8)&0xff, c->user_data);
      c->callback ((x3>>16)&0xff, c->user_data);
      c->callback ((x3>>24)&0xff, c->user_data);
      c->b = 0;
      c->d0 = 0;
      c->d1 = 0;
      c->d2 = 0;
      c->d3 = 0;
    }
}

void
Twofish_ecb_stop (struct twf_ecb_ctx *c)
{
  for ( ; c->b ; ) /* ``Cryptic'', isn't it? */
    Twofish_ecb_feed (c, 0);
}

void
Twofish_cbc_start (struct twf_cbc_ctx *c, char encrypt,
		    const unsigned char key[], int key_bytes,
		    void (* callback) (unsigned char byte, void *user_data),
		    void *user_data)
{
  Twofish_cbc_start_iv (c, encrypt, key, key_bytes, 0, 0, 0, 0,
			callback, user_data);
}

void
Twofish_cbc_start_iv (struct twf_cbc_ctx *c, char encrypt,
		      const unsigned char key[], int key_bytes,
		      unsigned long iv0, unsigned long iv1,
		      unsigned long iv2, unsigned long iv3,
		      void (* callback) (unsigned char byte, void *user_data),
		      void *user_data)
{
  c->encrypt = encrypt;
  Twofish_initialize (&c->c, key, key_bytes);
  c->b = 0;
  c->d0 = 0;  c->d1 = 0;  c->d2 = 0;  c->d3 = 0;
  c->l0 = iv0;  c->l1 = iv1;  c->l2 = iv2;  c->l3 = iv3;
  c->callback = callback;
  c->user_data = user_data;
}

void
Twofish_cbc_feed (struct twf_cbc_ctx *c, unsigned char inb)
{
  if ( c->b < 4 )
    c->d0 |= ((unsigned long)inb)<<(8*c->b);
  else if ( c->b < 8 )
    c->d1 |= ((unsigned long)inb)<<(8*(c->b-4));
  else if ( c->b < 12 )
    c->d2 |= ((unsigned long)inb)<<(8*(c->b-8));
  else
    c->d3 |= ((unsigned long)inb)<<(8*(c->b-12));
  c->b++;
  if ( c->b >= 16 )
    /* We have one block of data */
    {
      unsigned long x0, x1, x2, x3;

      x0 = c->d0;  x1 = c->d1;  x2 = c->d2;  x1 = c->d2;
      if ( c->encrypt )
	{
	  x0 ^= c->l0;  x1 ^= c->l1;  x2 ^= c->l2;  x1 ^= c->l2;
	}
      (c->encrypt ? Twofish_encipher : Twofish_decipher) (&c->c,
							  &x0, &x1, &x2, &x3);
      if ( c->encrypt )
	{
	  /* Keep encrypted block to encrypt next one */
	  c->l0 = x0;  c->l1 = x1;  c->l2 = x2;  c->l3 = x3;
	}
      else
	{
	  x0 ^= c->l0;  x1 ^= c->l1;  x2 ^= c->l2;  x3 ^= c->l3;
	  /* Keep last block of ciphertext to decrypt next one */
	  c->l0 = c->d0;  c->l1 = c->d1;  c->l2 = c->d2;  c->l3 = c->d3;
	}
      c->callback (x0&0xff, c->user_data);
      c->callback ((x0>>8)&0xff, c->user_data);
      c->callback ((x0>>16)&0xff, c->user_data);
      c->callback ((x0>>24)&0xff, c->user_data);
      c->callback (x1&0xff, c->user_data);
      c->callback ((x1>>8)&0xff, c->user_data);
      c->callback ((x1>>16)&0xff, c->user_data);
      c->callback ((x1>>24)&0xff, c->user_data);
      c->callback (x2&0xff, c->user_data);
      c->callback ((x2>>8)&0xff, c->user_data);
      c->callback ((x2>>16)&0xff, c->user_data);
      c->callback ((x2>>24)&0xff, c->user_data);
      c->callback (x3&0xff, c->user_data);
      c->callback ((x3>>8)&0xff, c->user_data);
      c->callback ((x3>>16)&0xff, c->user_data);
      c->callback ((x3>>24)&0xff, c->user_data);
      c->b = 0;
      c->d0 = 0;
      c->d1 = 0;
      c->d2 = 0;
      c->d3 = 0;
    }
}

void
Twofish_cbc_stop (struct twf_cbc_ctx *c)
{
  for ( ; c->b ; ) /* ``Cryptic'', isn't it? */
    Twofish_cbc_feed (c, 0);
}

void
Twofish_cfb_start (struct twf_cfb_ctx *c, char encrypt,
		   const unsigned char key[], int key_bytes,
		   void (* callback) (unsigned char byte, void *user_data),
		   void *user_data)
{
  Twofish_cfb_start_iv (c, encrypt, key, key_bytes, 0, 0, 0, 0,
			callback, user_data);
}

void
Twofish_cfb_start_iv (struct twf_cfb_ctx *c, char encrypt,
		      const unsigned char key[], int key_bytes,
		      unsigned long iv0, unsigned long iv1,
		      unsigned long iv2, unsigned long iv3,
		      void (* callback) (unsigned char byte, void *user_data),
		      void *user_data)
{
  c->encrypt = encrypt;
  Twofish_initialize (&c->c, key, key_bytes);
  c->b = 0;
  c->fr0 = iv0;  c->fr1 = iv1;  c->fr2 = iv2;  c->fr3 = iv3;
  c->callback = callback;
  c->user_data = user_data;
}

void
Twofish_cfb_feed (struct twf_cfb_ctx *c, unsigned char inb)
{
  unsigned char outb;

  if ( c->b == 0 )
    /* Starting a new block: encrypt the feedback register. */
    {
      unsigned long x0, x1, x2, x3;
      x0 = c->fr0;  x1 = c->fr1;  x2 = c->fr2;  x3 = c->fr3;
      Twofish_encipher (&c->c, &x0, &x1, &x2, &x3);
      c->fre0 = x0;  c->fre1 = x1;  c->fre2 = x2;  c->fre3 = x3;
#if 0 /* Don't do this as it breaks resync */
      c->fr0 = 0;  c->fr1 = 0;  c->fr2 = 0;  c->fr3 = 0;
#endif
    }
  if ( c->b < 4 )
    outb = inb ^ ((c->fre0>>(8*(c->b)))&0xff);
  else if ( c->b < 8 )
    outb = inb ^ ((c->fre1>>(8*(c->b-4)))&0xff);
  else if ( c->b < 12 )
    outb = inb ^ ((c->fre2>>(8*(c->b-8)))&0xff);
  else
    outb = inb ^ ((c->fre3>>(8*(c->b-12)))&0xff);
  c->fr0 = (c->fr0>>8) | ((c->fr1<<24)&0xffffffffUL);
  c->fr1 = (c->fr1>>8) | ((c->fr2<<24)&0xffffffffUL);
  c->fr2 = (c->fr2>>8) | ((c->fr3<<24)&0xffffffffUL);
  if ( c->encrypt )
    c->fr3 = (c->fr3>>8) | (((unsigned long)outb)<<24);
  else
    c->fr3 = (c->fr3>>8) | (((unsigned long)inb)<<24);
  c->callback (outb, c->user_data);
  c->b++;
  if ( c->b >= 16 )
    c->b = 0;
}

#if defined(__GNUC__)
void
Twofish_cfb_stop (struct twf_cfb_ctx *c __attribute__((__unused__)))
#else
void
Twofish_cfb_stop (struct twf_cfb_ctx *c)
#endif
{
  /* Nada */
}

void
Twofish_cfb_resync (struct twf_cfb_ctx *c)
{
  c->b = 0;
}
