/* Twofish routines, after Bruce Schneier &al */
/* Written by David Madore <david.madore@ens.fr>. */
/* Public domain (2001-05-11) */
/* This version last modified 2003-02-16 */

/* Note: these routines do not depend on endianness. */

/* === The header === */

/* Put this in twofish.h if you don't like having everything in one
 * big file. */

#ifndef _DMADORE_TWOFISH_H
#define _DMADORE_TWOFISH_H

/* --- Basic twofish routines --- */

#define TWOFISH_NBROUNDS 16

struct twf_ctx {
  /* The subkeys used by the twofish cipher */
  unsigned long K[2*TWOFISH_NBROUNDS+8];
  unsigned long S[4][256];
};

/* Encipher one 128-bit quantity (divided in four 32-bit quantities)
 * using the precalculated subkeys). */
void Twofish_encipher (const struct twf_ctx *c,
		       unsigned long *x0, unsigned long *x1,
		       unsigned long *x2, unsigned long *x3);

/* Decipher one 128-bit quantity (divided in four 32-bit quantities)
 * using the precalculated subkeys). */
void Twofish_decipher (const struct twf_ctx *c,
		       unsigned long *x0, unsigned long *x1,
		       unsigned long *x2, unsigned long *x3);

/* Initialize the cipher by calculating the subkeys from the key. */
void Twofish_initialize (struct twf_ctx *c,
			 const unsigned char key[], int key_bytes);

/* --- Twofish used in Electronic Code Book (ECB) mode --- */

struct twf_ecb_ctx {
  /* Whether we are encrypting (rather than decrypting) */
  char encrypt;
  /* The twofish subkeys */
  struct twf_ctx c;
  /* The 128 bits of data being written */
  unsigned long d0, d1, d2, d3;
  /* Our position within the 128 bits (always between 0 and 15) */
  int b;
  /* The callback function to be called with every byte produced */
  void (* callback) (unsigned char byte, void *user_data);
  /* The user data to pass the the callback function */
  void *user_data;
};

/* Start an ECB Twofish cipher session: specify whether we are
 * encrypting or decrypting, what key is to be used, and what callback
 * should be called for every byte produced. */
void Twofish_ecb_start (struct twf_ecb_ctx *c, char encrypt,
			const unsigned char key[], int key_bytes,
			void (* callback) (unsigned char byte,
					   void *user_data),
			void *user_data);

/* Feed one byte to an ECB Twofish cipher session. */
void Twofish_ecb_feed (struct twf_ecb_ctx *c, unsigned char inb);

/* Stop an ECB Twofish session (i.e. flush the remaining bytes). */
void Twofish_ecb_stop (struct twf_ecb_ctx *c);

/* --- Twofish used in Cipher Block Chaining (CBC) mode --- */

struct twf_cbc_ctx {
  /* Whether we are encrypting (rather than decrypting) */
  char encrypt;
  /* The twofish subkeys */
  struct twf_ctx c;
  /* The 128 bits of data being written */
  unsigned long d0, d1, d2, d3;
  /* The last 128 bits of ciphertext */
  unsigned long l0, l1, l2, l3;
  /* Our position within the 128 bits (always between 0 and 15) */
  int b;
  /* The callback function to be called with every byte produced */
  void (* callback) (unsigned char byte, void *user_data);
  /* The user data to pass the the callback function */
  void *user_data;
};

/* Start a CBC Twofish cipher session: specify whether we are
 * encrypting or decrypting, what key is to be used, and what callback
 * should be called for every byte produced. */
void Twofish_cbc_start (struct twf_cbc_ctx *c, char encrypt,
			const unsigned char key[], int key_bytes,
			void (* callback) (unsigned char byte,
					   void *user_data),
			void *user_data);

/* Ditto, specifying Initialization Vector. */
void Twofish_cbc_start_iv (struct twf_cbc_ctx *c, char encrypt,
			   const unsigned char key[], int key_bytes,
			   unsigned long iv0, unsigned long iv1,
			   unsigned long iv2, unsigned long iv3,
			   void (* callback) (unsigned char byte,
					      void *user_data),
			   void *user_data);

/* Feed one byte to a CBC Twofish cipher session. */
void Twofish_cbc_feed (struct twf_cbc_ctx *c, unsigned char inb);

/* Stop a CBC Twofish session (i.e. flush the remaining bytes). */
void Twofish_cbc_stop (struct twf_cbc_ctx *c);

/* --- Twofish used in Cipher Feedback (CFB) mode --- */

struct twf_cfb_ctx {
  /* Whether we are encrypting (rather than decrypting) */
  char encrypt;
  /* The twofish subkeys */
  struct twf_ctx c;
  /* The 128-bit feedback register */
  unsigned long fr0, fr1, fr2, fr3;
  /* The 128-bit encrypted feedback register */
  unsigned long fre0, fre1, fre2, fre3;
  /* Our position within the 128 bits (always between 0 and 15) */
  int b;
  /* The callback function to be called with every byte produced */
  void (* callback) (unsigned char byte, void *user_data);
  /* The user data to pass the the callback function */
  void *user_data;
};

/* Start a CFB Twofish cipher session: specify whether we are
 * encrypting or decrypting, what key is to be used, and what callback
 * should be called for every byte produced. */
void Twofish_cfb_start (struct twf_cfb_ctx *c, char encrypt,
			const unsigned char key[], int key_bytes,
			void (* callback) (unsigned char byte,
					   void *user_data),
			void *user_data);

/* Ditto, specifying Initialization Vector. */
void Twofish_cfb_start_iv (struct twf_cfb_ctx *c, char encrypt,
			   const unsigned char key[], int key_bytes,
			   unsigned long iv0, unsigned long iv1,
			   unsigned long iv2, unsigned long iv3,
			   void (* callback) (unsigned char byte,
					      void *user_data),
			   void *user_data);

/* Feed one byte to a CFB Twofish cipher session. */
void Twofish_cfb_feed (struct twf_cfb_ctx *c, unsigned char inb);

/* Stop a CFB Twofish session (i.e. flush the remaining bytes). */
void Twofish_cfb_stop (struct twf_cfb_ctx *c);

/* "Resync" a CFB Twofish session - this is used in OpenPGP (see RFC
 * 2440 section 12.8). */
void Twofish_cfb_resync (struct twf_cfb_ctx *c);

#endif /* not defined _DMADORE_TWOFISH_H */

/* EOF */
